#include "chainerx/kernels/rnn.h"

namespace chainerx {
namespace internal {
CHAINERX_REGISTER_BUILTIN_KEY_KERNEL(Rnn)
CHAINERX_REGISTER_BUILTIN_KEY_KERNEL(RnnBackward)
}  // namespace internal
}  // namespace chainerx
